﻿using System;
using System.Drawing;
using System.Globalization;
using System.IO;
using System.Threading;
using System.Windows;
using System.Windows.Forms;
using System.Windows.Forms.DataVisualization.Charting;
using MessageBox = System.Windows.MessageBox;

namespace ChartSample
{
    /// <summary>
    /// Логика взаимодействия для MainWindow.xaml
    /// </summary>
    public partial class MainWindow
    {

        private Chart _chartForCandle; // это хранится адрес для нашего чарта

        private string _pathToHistory; // это адрес для строки с путём к  данным

        private Candle[] _candleArray; // это массив свечек

        public MainWindow() //окно загружается.
        {
            InitializeComponent();
            // вызываем метод для создания чарта
            CreateChart();
        }

        private void CreateChart() // метод создающий чарт
        {
            // создаём чарт от Win Forms
            _chartForCandle = new Chart();
            // привязываем его к хосту. 
            ChartHost.Child = _chartForCandle;
            ChartHost.Child.Show();

            // на всякий случай чистим в нём всё
            _chartForCandle.Series.Clear();
            _chartForCandle.ChartAreas.Clear();

// свечи

            ChartArea candleArea = new ChartArea("ChartAreaCandle") // создаём область на графике
            {
                CursorX = {IsUserSelectionEnabled = true, IsUserEnabled = true},
                // разрешаем пользователю изменять рамки представления
                CursorY = {AxisType = AxisType.Secondary},
                Position = {Height = 70, X = 0, Width = 100, Y = 0}, //чертa
            };

            _chartForCandle.ChartAreas.Add(candleArea); // добавляем область на чарт



            Series candleSeries = new Series("SeriesCandle") // создаём для нашей области коллекцию значений
            {
                ChartType = SeriesChartType.Candlestick, // назначаем этой коллекции тип "Свечи"
                YAxisType = AxisType.Secondary,// назначаем ей правую линейку по шкале Y (просто для красоты) Везде ж так
                ChartArea = "ChartAreaCandle", // помещаем нашу коллекцию на ранее созданную область
                ShadowOffset = 2 // наводим тень
            };

            _chartForCandle.Series.Add(candleSeries); // добавляем коллекцию на чарт

// объём
            ChartArea volumeArea = new ChartArea("ChartAreaVolume") // создаём область для объёма
            {
                CursorX = {IsUserEnabled = true}, //чертa
                CursorY = {AxisType = AxisType.Secondary}, // ось У правая
                AlignWithChartArea = "ChartAreaCandle",// выравниваем по верхней диаграмме
                Position = {Height = 30, X = 0, Width = 100, Y = 70},
                AxisX = { Enabled = AxisEnabled.False }// отменяем прорисовку оси Х
            };

            _chartForCandle.ChartAreas.Add(volumeArea);

            Series volumeSeries = new Series("SeriesVolume") // создаём для нашей области коллекцию значений
            {
                ChartType = SeriesChartType.Column, // назначаем этой коллекции тип "столбцы"
                YAxisType = AxisType.Secondary, // назначаем ей правую линейку по шкале Y (просто для красоты)
                ChartArea = "ChartAreaVolume", // помещаем нашу коллекцию на ранее созданную область
                ShadowOffset = 2 // наводим тень на плетень
            };

            _chartForCandle.Series.Add(volumeSeries);

            // общее
            foreach (ChartArea area in _chartForCandle.ChartAreas)
            {
                // Делаем курсор по Y красным и толстым
                area.CursorX.LineColor = Color.Red;
                area.CursorX.LineWidth = 2;
            }

            // подписываемся на события изменения масштабов
            _chartForCandle.AxisScrollBarClicked += chartForCandle_AxisScrollBarClicked; // событие передвижения курсора
            _chartForCandle.AxisViewChanged += chartForCandle_AxisViewChanged; // событие изменения масштаба
            _chartForCandle.CursorPositionChanged += chartForCandle_CursorPositionChanged;
            // событие выделения диаграммы
        }

        private void ButtonRew_Click(object sender, RoutedEventArgs e) // кнопка перемотать
        {
            _chartForCandle.Visible = false; // прячем чарт
        }

        private void ButtonSendHistory_Click(object sender, RoutedEventArgs e) // кнопка указать историю
        {
            OpenFileDialog myDialog = new OpenFileDialog
            {
                // создаём стандартное спецОкошко для указания файла с историей
                CheckFileExists = true,
                Multiselect = false
            };

            myDialog.ShowDialog();

            if (myDialog.FileName != "") // если хоть что-то выбрано и это свечи
            {
                // здесь происходит сохранение адреса выбранного фала.
                // по хорошему надо бы здесь поставить проверку, что в нём лежит
                _pathToHistory = myDialog.FileName;
            }
        }

        private void ButtonSlowStart_Click(object sender, RoutedEventArgs e) // кнопка медленная прорисовка
        {
            // медленная прорисовка графика, по одной свечке

            if (_pathToHistory == null)
            {
                // если история ещё не подключена
                MessageBox.Show("Прежде чем прогрузить график, надо подгрузить историю");
                return;
            }

            foreach (Series series in _chartForCandle.Series)
            {
                // очищаем все свечки на графике
                series.Points.Clear();
            }

            ChartArea candleArea = _chartForCandle.ChartAreas.FindByName("ChartAreaCandle");
            while (candleArea != null && candleArea.AxisX.ScaleView.IsZoomed)
            {
                // сбрасываем увеличение на чарте
                candleArea.AxisX.ScaleView.ZoomReset();
            }

            ButtonFastStart.IsEnabled = false;
            ButtonSlowStart.IsEnabled = false;
            ButtonRew.IsEnabled = true;

            // вызываем метод прорисовки графика, но делаем это отдельным потоком, чтобы форма оставалась живой
            Thread worker = new Thread(StartPaintChartSlow) { IsBackground = true };
            worker.Start();
        }

        private void StartPaintChartSlow() // метод вызывающийся в новом потоке, для прорисовки графика
        {
            LoadCandleFromFile();
            LoadCandleOnChartSlow();
        }

        private void LoadCandleFromFile() // загрузить свечки из файла
        {
            try
            {
                // используем перехватчик исключений, т.к. файл может быть занят или содержать каку.

                int lenghtArray = 0;

                using (StreamReader reader = new StreamReader(_pathToHistory))
                {
                    // подсоединяемся к файлу
                    while (!reader.EndOfStream)
                    {
                        //считаем кол-во строк
                        lenghtArray++;
                        reader.ReadLine();
                    }
                }

                Candle[] newCandleArray = new Candle[lenghtArray];


                using (StreamReader reader = new StreamReader(_pathToHistory))
                {
                    // подсоединяемся к файлу

                    for (int iteratorArray = 0; iteratorArray < newCandleArray.Length; iteratorArray++)
                    {
                        // закачиваем свечки из файла в массив
                        newCandleArray[iteratorArray] = new Candle();
                        newCandleArray[iteratorArray].SetCandleFromString(reader.ReadLine());
                    }
                }

                _candleArray = newCandleArray; // сохраняем изменения
            }
            catch (Exception error)
            {
                MessageBox.Show("Произошла ошибка при скачивании данных из файла. Ошибка: " + error);
            }
        }

        private void LoadCandleOnChartSlow() // прогрузить загруженные свечки на график по одной
        {

            if (_candleArray == null)
            {
                //если наш массив пуст по каким-то причинам
                return;
            }

            for (int i = 0; i < _candleArray.Length; i++)
            {
                // отправляем наш массив по свечкам на прорисовку
                Thread.Sleep(1);
                // спим 5ть миллисекунд между свечками, чтобы форма не висела и могла отвечать на запросы пользователя 
                LoadNewCandle(_candleArray[i], i);
            }

            StopLoadHistory();

        }

        private void LoadNewCandle(Candle newCandle, int numberInArray) // добавить одну свечу на график
        {
            if (!CheckAccess())
            {
                // перезаходим в метод потоком формы, чтобы не было исключения
                Dispatcher.Invoke(new Action<Candle, int>(LoadNewCandle), newCandle, numberInArray);
                return;
            }
            // свечи
            Series candleSeries = _chartForCandle.Series.FindByName("SeriesCandle");

            if (candleSeries != null)
            {
                // забиваем новую свечку
                candleSeries.Points.AddXY(numberInArray, newCandle.Low, newCandle.High, newCandle.Open, newCandle.Close);

                // подписываем время
                candleSeries.Points[candleSeries.Points.Count - 1].AxisLabel =
                    newCandle.Time.ToString(CultureInfo.InvariantCulture);

                // разукрышиваем в привычные цвета
                if (newCandle.Close > newCandle.Open)
                {
                    candleSeries.Points[candleSeries.Points.Count - 1].Color = Color.Green;
                }
                else
                {
                    candleSeries.Points[candleSeries.Points.Count - 1].Color = Color.Red;
                }

                ChartArea candleArea = _chartForCandle.ChartAreas.FindByName("ChartAreaCandle");
                if (candleArea != null && candleArea.AxisX.ScrollBar.IsVisible) // если уже выбран какой-то диапазон
                {
                    // сдвигаем представление вправо
                    candleArea.AxisX.ScaleView.Scroll(_chartForCandle.ChartAreas[0].AxisX.Maximum);
                }
            }
            // объём
            Series volumeSeries = _chartForCandle.Series.FindByName("SeriesVolume");

            if (volumeSeries != null)
            {
                volumeSeries.Points.AddXY(numberInArray, newCandle.Volume);
                // разукрышиваем в привычные цвета
                if (volumeSeries.Points.Count > 1)
                {
                    if (volumeSeries.Points[volumeSeries.Points.Count - 2].YValues[0] < newCandle.Volume)
                    {
                        volumeSeries.Points[volumeSeries.Points.Count - 1].Color = Color.Green;
                    }
                    else
                    {
                        volumeSeries.Points[volumeSeries.Points.Count - 1].Color = Color.Red;
                    }
                }
            }

            ChartResize(); // Выводим нормальные рамки
        }

        private void ChartResize() // устанавливает границы представления по оси У
        {
            // вообще-то можно это автоматике доверить, но там вечно косяки какие-то, поэтому лучше самому следить за всеми осями
            try
            {
                if (_candleArray == null)
                {
                    return;
                }
                // свечи
                Series candleSeries = _chartForCandle.Series.FindByName("SeriesCandle");
                ChartArea candleArea = _chartForCandle.ChartAreas.FindByName("ChartAreaCandle");

                if (candleArea == null ||
                    candleSeries == null)
                {
                    return;
                }

                int startPozition = 0; // первая отображаемая свеча
                int endPozition = candleSeries.Points.Count; // последняя отображаемая свеча

                if (_chartForCandle.ChartAreas[0].AxisX.ScrollBar.IsVisible)
                {
                    // если уже выбран какой-то диапазон, назначаем первую и последнюю исходя из этого диапазона

                    startPozition = Convert.ToInt32(candleArea.AxisX.ScaleView.Position);
                    endPozition = Convert.ToInt32(candleArea.AxisX.ScaleView.Position) +
                                  Convert.ToInt32(candleArea.AxisX.ScaleView.Size);
                }

                candleArea.AxisY2.Maximum = GetMaxValueOnChart(_candleArray, startPozition, endPozition);
                candleArea.AxisY2.Minimum = GetMinValueOnChart(_candleArray, startPozition, endPozition);

                // объёмы
                Series volumeSeries = _chartForCandle.Series.FindByName("SeriesVolume");
                ChartArea volumeArea = _chartForCandle.ChartAreas.FindByName("ChartAreaVolume");

                if (volumeSeries != null &&
                    volumeArea != null)
                {
                    volumeArea.AxisY2.Maximum = GetMaxVolume(_candleArray, startPozition, endPozition);
                    volumeArea.AxisY2.Minimum = 0;
                }

                _chartForCandle.Refresh();
            }
            catch (Exception error)
            {
                MessageBox.Show("Обибка при изменении ширины представления. Ошибка: " + error);
            }
        }

        private double GetMaxVolume(Candle[] book, int start, int end) // берёт максимальное значение объёма за период
        {
            double result = double.MinValue;

            for (int i = start; i < end && i < book.Length; i++)
            {
                if (book[i].Volume > result)
                {
                    result = book[i].Volume;
                }
            }

            return result;
        }

        private double GetMinValueOnChart(Candle[] book, int start, int end)
        // берёт минимальное значение из массива свечек
        {
            double result = double.MaxValue;

            for (int i = start; i < end && i < book.Length; i++)
            {
                if (book[i].Low < result)
                {
                    result = book[i].Low;
                }
            }

            return result;
        }

        private double GetMaxValueOnChart(Candle[] book, int start, int end)
        // берёт максимальное значение из массива свечек
        {
            double result = 0;

            for (int i = start; i < end && i < book.Length; i++)
            {
                if (book[i].High > result)
                {
                    result = book[i].High;
                }
            }

            return result;
        }

        private void StopLoadHistory() // вызываем при завершении прорисовки
        {
            if (!CheckAccess())
            {
                // перезаходим в метод потоком формы, чтобы не было исключения
                Dispatcher.Invoke(new Action(StopLoadHistory));
                return;
            }

            if (_chartForCandle.Visible == false)
            {
                _chartForCandle.Visible = true;
                _chartForCandle.Refresh();
            }

            ButtonFastStart.IsEnabled = true;
            ButtonSlowStart.IsEnabled = true;
            ButtonRew.IsEnabled = false;
        }

        private void ButtonFastStart_Click(object sender, RoutedEventArgs e) // вызывает быструю прорисовку
        {
            // медленная прорисовка графика, по одной свечке

            if (_pathToHistory == null)
            {
                // если история ещё не подключена
                MessageBox.Show("Прежде чем прогрузить график, надо подгрузить историю");
                return;
            }

            foreach (Series series in _chartForCandle.Series)
            {
                // очищаем все свечки на графике
                series.Points.Clear();
            }

            ChartArea candleArea = _chartForCandle.ChartAreas.FindByName("ChartAreaCandle");
            while (candleArea != null && candleArea.AxisX.ScaleView.IsZoomed)
            {
                // сбрасываем увеличение на чарте
                candleArea.AxisX.ScaleView.ZoomReset();
            }

            ButtonFastStart.IsEnabled = false;
            ButtonSlowStart.IsEnabled = false;

            // вызываем метод прорисовки графика, но делаем это отдельным потоком, чтобы форма оставалась живой
            Thread worker = new Thread(FastStart) { IsBackground = true };
            worker.Start();
        }

        private void FastStart()
        {
            LoadCandleFromFile();
            LoadCandleOnChartFast();
        }

        private void LoadCandleOnChartFast() // формирует серии данных
        {
            // суть быстрой прогрузки в том, чтобы создать уже готовые серии данных и только потом
            // подгружать их на график. 
            Series candleSeries = new Series("SeriesCandle")
            {
                ChartType = SeriesChartType.Candlestick,// назначаем этой коллекции тип "Свечи"
                YAxisType = AxisType.Secondary,// назначаем ей правую линейку по шкале Y (просто для красоты)
                ChartArea = "ChartAreaCandle",// помещаем нашу коллекцию на ранее созданную область
                ShadowOffset = 2,  // наводим тень
                YValuesPerPoint = 4 // насильно устанавливаем число У точек для серии
            };

            for (int i = 0; i < _candleArray.Length; i++)
            {
                // забиваем новую свечку
                candleSeries.Points.AddXY(i, _candleArray[i].Low, _candleArray[i].High, _candleArray[i].Open,
                    _candleArray[i].Close);

                // подписываем время
                candleSeries.Points[candleSeries.Points.Count - 1].AxisLabel =
                    _candleArray[i].Time.ToString(CultureInfo.InvariantCulture);

                // разукрышиваем в привычные цвета
                if (_candleArray[i].Close > _candleArray[i].Open)
                {
                    candleSeries.Points[candleSeries.Points.Count - 1].Color = Color.Green;
                }
                else
                {
                    candleSeries.Points[candleSeries.Points.Count - 1].Color = Color.Red;
                }
            }

            Series volumeSeries = new Series("SeriesVolume")
            {
                ChartType = SeriesChartType.Column, // назначаем этой коллекции тип "Свечи"
                YAxisType = AxisType.Secondary,// назначаем ей правую линейку по шкале Y (просто для красоты) Везде ж так
                ChartArea = "ChartAreaVolume", // помещаем нашу коллекцию на ранее созданную область
                ShadowOffset = 2 // наводим тень
            };

            for (int i = 0; i < _candleArray.Length; i++)
            {
                volumeSeries.Points.AddXY(i, _candleArray[i].Volume);
                // разукрышиваем в привычные цвета
                if (volumeSeries.Points.Count > 1)
                {
                    if (volumeSeries.Points[volumeSeries.Points.Count - 2].YValues[0] < _candleArray[i].Volume)
                    {
                        volumeSeries.Points[volumeSeries.Points.Count - 1].Color = Color.Green;
                    }
                    else
                    {
                        volumeSeries.Points[volumeSeries.Points.Count - 1].Color = Color.Red;
                    }
                }
            }

            SetSeries(candleSeries, volumeSeries);
            StopLoadHistory();
        }

        private void SetSeries(Series candleSeries, Series volumeSeries) // подгружает серии данных на график
        {
            if (!CheckAccess())
            {
                // перезаходим в метод потоком формы, чтобы не было исключения
                Dispatcher.Invoke(new Action<Series, Series>(SetSeries), candleSeries, volumeSeries);
                return;
            }

            _chartForCandle.Series.Clear(); // убираем с нашего графика все до этого созданные серии с данными

            _chartForCandle.Series.Add(candleSeries);
            _chartForCandle.Series.Add(volumeSeries);

            ChartArea candleArea = _chartForCandle.ChartAreas.FindByName("ChartAreaCandle");
            if (candleArea != null && candleArea.AxisX.ScrollBar.IsVisible)
            // если уже выбран какой-то диапазон
            {
                // сдвигаем представление вправо
                candleArea.AxisX.ScaleView.Scroll(_chartForCandle.ChartAreas[0].AxisX.Maximum);
            }
            ChartResize();
            _chartForCandle.Refresh();
        }

        // события
        private void chartForCandle_CursorPositionChanged(object sender, CursorEventArgs e)
        // событие изменение отображения диаграммы
        {
            ChartResize();
        }

        private void chartForCandle_AxisViewChanged(object sender, ViewEventArgs e)
        // событие изменение отображения диаграммы 
        {
            ChartResize();
        }

        private void chartForCandle_AxisScrollBarClicked(object sender, ScrollBarEventArgs e)
        // событие изменение отображения диаграммы
        {
            ChartResize();
        }
    }

}
